//
// anon on 21.03.23.
//

#include "create_training_data.h"

#include <string>
#include <vector>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <thread>

#include <util/replace_symbol.h>
#include <util/replace_expr.h>
#include <util/cmdline.h>
#include <util/cout_message.h>
#include <util/format_expr.h>
#include <util/namespace.h>
#include <util/symbol_table.h>
#include <util/std_expr.h>
#include <util/expr.h>
#include <util/expr_iterator.h>
#include <solvers/smt2/smt2_dec.h>
#include <unordered_set>


#include "problem.h"
#include "term_position.h"
#include "anti_unification.h"
#include "unification.h"
#include "constants.h"
#include "progressbar.hpp"
#include "sygus_problem.h"
#include "smt2_frontend.h"
#include "parser.h"
#include "util.h"
#include "subterm_heuristics.h"
#include "expr2sygus.h"
#include "cvc5_synth.h"


std::vector<exprt> symbol_expr_vec_to_expr_vec(const std::vector<symbol_exprt>& vec) {
    std::vector<exprt> new_vars;
    new_vars.reserve(vec.size());
    for (const auto &x: vec) {
        new_vars.emplace_back(x);
    }

    return new_vars;
}

/**
 * We Generate the substitutions with the synth fun target for all of the terms and the LGG:
 * if the terms are 4 + 6, 1 + 2
 * with LGG x + y,
 * application term synthTarget(x,y)
 * we generate the subst map {4+6 -> synthTarget(4,6), synthTarget(1,2)}.
 *
 * */
replace_mapt get_substitution_for_SO_var(const std::vector<exprt>& terms, const exprt& lgg, const function_application_exprt& app) {
    replace_mapt replace_map;
    for (auto &term: terms) {
        std::vector<std::pair<exprt, exprt>> to_unify;
        // The order of arguments here is important due to the ordering in the if-else-if branches in unify.
        // Note that this should be solved at some point with variable ordering. But until then let's keep it this way.
        to_unify.emplace_back(lgg, term);
        replace_symbolt substitution = unify(to_unify).value();
        exprt tmp = app;
        substitution(tmp);
        // replace term with tmp in smt_problem
        replace_map.insert({term, tmp});
    }

    return replace_map;
}


std::string map_logic_from_smt_to_sygus(std::string basicString) {
    if (basicString.rfind("QF_",0) == 0) { // check if it starts with
        return basicString.substr(3);
    } else {
        return basicString;
    }
}

std::optional<sygus_problemt> create_cvc5_adversarial_problem(const smt_problemt &smt_problem,
                                                              const namespacet &namespacet,
                                                              message_handlert& mh) {

    int lgg_max_height = 100;
    std::vector<exprt> relevant_subterms = get_many_subterms_filtered(smt_problem);

    std::set<std::pair<exprt,exprt>> done;

    for (int i = 0; i < relevant_subterms.size(); ++i) {
        const auto &x = relevant_subterms[i];
        for (int j = i+1; j < relevant_subterms.size(); ++j) {
            const auto &y = relevant_subterms[j];


            if (done.find(std::pair<exprt, exprt>{x, y}) != done.end()) {
                continue;
            }
            if (!relevant_term_pair(x,y)) {
                continue;
            }

            auto lgg = compute_lgg({{x, y}});
            if (!is_binder_free(lgg.first)) {
                continue;
            }
            auto lgg_height = expr_height(lgg.first);

            std::vector<exprt> two_terms{x, y};
            for (const auto &term: relevant_subterms) {
                if (term.type() != x.type()) {
                    continue;
                }
                if (!is_binder_free(term)) {
                    continue;
                }

                two_terms.push_back(term);
                auto new_lgg = compute_lgg(two_terms);
                if (expr_height(new_lgg.first) < lgg_height) {
                    two_terms.pop_back();
                } else {
                    lgg = new_lgg;
                }
            }

            done.emplace(x, y);


            std::vector<term_positiont> positions;

            for (const auto & assertion : smt_problem.assertions) {
                for (const auto &t: two_terms) {
                    concat(positions, get_pos_of_all_occurrences(t, assertion, term_positiont(i)));
                }
            }

            if (positions.empty()) {
                return std::nullopt;
            }

            std::vector<exprt> terms;
            terms.reserve(positions.size());
            for (auto &pos: positions) {
                terms.push_back(get_term_copy_at_position_in_problem(pos, smt_problem));
            }

            std::vector<exprt> new_vars = symbol_expr_vec_to_expr_vec(lgg.second);

            // this is a term like :    Foo(t1,...,tn) where n >= number of variables gathered from lgg;
            // exprt second_order_var("synthTarget", lgg.type(), std::vector<exprt>(new_vars));
            function_application_exprt synth_fun_app =
                    create_func_app("synthTarget", std::vector<exprt>(new_vars), lgg.first.type());

            // generate the replace map
            replace_mapt replace_map = get_substitution_for_SO_var(terms, lgg.first, synth_fun_app);

            // Creating Resulting SyGuS problem
            sygus_problemt sygus_problem;
            // add comments
            sygus_problem.comments = smt_problem.comments;
            std::stringstream ss;
            ss << "Solution: " << format(lgg.first);
            sygus_problem.comments.push_back(ss.str());
            //add synth func command
            synth_fun_commandt synth_fun;
            synth_fun.id = to_symbol_expr(synth_fun_app.function()).get_identifier();
            synth_fun.type = synth_fun_app.function_type();
            // synth_fun.parameters = irep_args;
            for (const auto &var: lgg.second) {
                synth_fun.parameters.push_back(var.get_identifier());
            }
            sygus_problem.synth_fun = synth_fun;
            // other stuff
            sygus_problem.logic = map_logic_from_smt_to_sygus(smt_problem.logic);
            sygus_problem.defined_functions = smt_problem.defined_functions;
            sygus_problem.filename = smt_problem.filename;
            // adding universally quantified variables
            for (const auto &free_var: smt_problem.free_var) {
                sygus_problem.free_var.push_back(free_var.first);
            }

            // replace in assertions
            for (auto assertion: smt_problem.assertions) {
                replace_expr(replace_map, assertion);
                if (simplify_expr(assertion,namespacet).is_true()) { // skip trivial constraints
                    continue;
                }
                sygus_problem.assertions.push_back(assertion);
            }

            cvc5_syntht cvc5_synth(mh,10000);
            decision_proceduret::resultt res = cvc5_synth(sygus_problem);
            if (res == decision_proceduret::resultt::D_SATISFIABLE ) {
                std::cout << "Synthesis succeeded" << std::endl;
                auto f = *cvc5_synth.get_solution().begin(); // we only have one symbol
                std::cout << format(f.first) << " = " << format(f.second)<<std::endl;
                if (expr_height(f.second) > 3) {
                    return {sygus_problem};
                }
            } else if (res == decision_proceduret::resultt::D_ERROR) {
                return {sygus_problem};
            }
        }
    }

    return std::nullopt;
}

std::optional<sygus_problemt> default_sygus_generation(const smt_problemt &smt_problem, const namespacet &namespacet) {

    std::vector<term_positiont> positions = get_term_positions(smt_problem);
    if (positions.empty()) {
        return std::nullopt;
    }
    std::vector<exprt> terms;
    terms.reserve(positions.size());
    for (auto &pos: positions) {
        terms.push_back(get_term_copy_at_position_in_problem(pos, smt_problem));
    }

    std::pair<exprt, std::vector<symbol_exprt>> anti_uni = compute_lgg(terms);

    exprt lgg = anti_uni.first;
    std::vector<exprt> new_vars = symbol_expr_vec_to_expr_vec(anti_uni.second);

    // this is a term like :    Foo(t1,...,tn) where n >= number of variables gathered from lgg;
    // exprt second_order_var("synthTarget", lgg.type(), std::vector<exprt>(new_vars));
    function_application_exprt synth_fun_app =
            create_func_app("synthTarget", std::vector<exprt>(new_vars), lgg.type());

    // generate the replace map
    replace_mapt replace_map = get_substitution_for_SO_var(terms, lgg, synth_fun_app);

    // Creating Resulting SyGuS problem
    sygus_problemt sygus_problem;
    // add comments
    sygus_problem.comments = smt_problem.comments;
    std::stringstream ss;
    ss << "Solution: " << format(lgg);
    sygus_problem.comments.push_back(ss.str());
    //add synth func command
    synth_fun_commandt synth_fun;
    synth_fun.id = to_symbol_expr(synth_fun_app.function()).get_identifier();
    synth_fun.type = synth_fun_app.function_type();
    // synth_fun.parameters = irep_args;
    for (const auto &x: anti_uni.second) {
        synth_fun.parameters.push_back(x.get_identifier());
    }
    sygus_problem.synth_fun = synth_fun;
    // other stuff
    sygus_problem.logic = map_logic_from_smt_to_sygus(smt_problem.logic);
    sygus_problem.defined_functions = smt_problem.defined_functions;
    sygus_problem.filename = smt_problem.filename;
    // adding universally quantified variables
    for (const auto &x: smt_problem.free_var) {
        sygus_problem.free_var.push_back(x.first);
    }

    // replace in assertions
    for (auto assertion: smt_problem.assertions) {
        replace_expr(replace_map, assertion);
        if (simplify_expr(assertion,namespacet).is_true()) { // skip trivial constraints
            continue;
        }
        sygus_problem.assertions.push_back(assertion);
    }

    return sygus_problem;
}


smt_problemt get_smt_problem_from_file(const std::string &file, messaget& message) {
    std::ifstream in(file);

    if (!in) {
        std::cerr << "Failed to open input file" << std::endl;
        throw default_exception("Failed to open file " + file);
    }

    in.ignore( std::numeric_limits<std::streamsize>::max() );
    std::streamsize length = in.gcount();
    in.clear();   //  Since ignore will have set eof.
    in.seekg( 0, std::ios_base::beg );
    if (length > MAX_FILE_SIZE) { // dirty hack to discard files that are too large and would lead to seg faults etc.
        throw std::exception();
    }

    parsert parser(in);
    // parse the smt_problem
    try {
        parser.parse();
    } catch (const parsert::smt2_errort &e) {
        message.error() << e.get_line_no() << ": "
                        << e.what() << messaget::eom;
        std::cout << "Could not parse: " << file << std::endl;
        throw std::exception();
        //throw default_exception("Could not parse.");
    }

    smt_problemt smt_problem = build_problem(parser);
    smt_problem.filename = file;
    expand_let_expressions(smt_problem);

    return smt_problem;
}


void write_problem(const sygus_problemt& s_prob, const std::string& file) {
    std::filesystem::path file_path(file);
    std::filesystem::create_directories(file_path.parent_path()); // create intermediate dirs.
    std::cout << "Writing result to: " << file_path << std::endl;

    std::ofstream myfile(file_path);
    myfile << build_sygus_query(s_prob);
    myfile.close();
}


int create_synthesis_problem(const cmdlinet &cmdline) {

    if (cmdline.args.size() < 2){
        return -1;
    }
    std::string input = cmdline.args[0];
    std::string output_file = cmdline.args[1];

    console_message_handlert message_handler;
    messaget message(message_handler);

    // this is our default verbosity
    unsigned int v = messaget::M_STATISTICS;

    if (cmdline.isset("verbosity")) {
        v = std::stol(cmdline.get_value("verbosity"));

        if (v > 10) {
            v = 10;

        }
    }


    message_handler.set_verbosity(v);

    auto smt_problem = get_smt_problem_from_file(input, message);

    symbol_tablet symbol_table;
    namespacet ns(symbol_table);
    smt_problemt new_valid_problem = generate_valid_problem(smt_problem, ns, message);

    std::optional<sygus_problemt> sygus_problem;

    if (!cmdline.isset("strategy")) {
        sygus_problem =  default_sygus_generation(new_valid_problem, ns);
    } else {
        // string strategy =  cmdline.get_value("strategy") TODO why does that not work?
        if (true) {
            sygus_problem = create_cvc5_adversarial_problem(new_valid_problem, ns, message_handler);
        }
    }

    if (sygus_problem) {
        write_problem(sygus_problem.value(), output_file);
    } else {
        std::cout << "Could not create problem." << std::endl;
    }

    return 0;
}