/*******************************************************************\

Module: SMT2 frontend example

Author: anon

\*******************************************************************/

#include "parser.h"
#include "printing_utils.h"
#include "problem.h"
#include "sygus_problem.h"
#include "cvc5_synth.h"
#include "util.h"

#include <fstream>
#include <iostream>

#include <util/cmdline.h>
#include <util/cout_message.h>
#include <util/format_expr.h>
#include <util/namespace.h>
#include <util/replace_symbol.h>
#include <util/simplify_expr.h>
#include <util/symbol_table.h>
#include <util/std_expr.h>
#include <util/expr.h>

#include <solvers/smt2/smt2_dec.h>

#include "constants.h"

#include "smt2_frontend.h"



// I would do any preprocessing in this function
smt_problemt build_problem(parsert &parser) {
    smt_problemt result;
    result.logic = parser.logic;
    for (const auto &a: parser.assertions) {
        result.assertions.push_back(a);
    }
    for (const auto &cmd: parser.set_info_cmds) {
        result.comments.push_back(cmd);
    }

    // identify the free variables and defined functions in the problem
    for (const auto &id: parser.id_map) {
        if (id.second.definition.is_nil() && id.second.kind == smt2_parsert::idt::VARIABLE) {
            result.free_var[symbol_exprt(id.first, id.second.type)] = nil_exprt();
        } else {
            result.defined_functions[symbol_exprt(id.first, id.second.type)] = id.second.definition;
        }
    }
    return result;
}


void expand_function_applications(exprt &expr, const smt_problemt &problem) {
    for (exprt &op: expr.operands()) {
        expand_function_applications(op, problem);
    }
    if (expr.id() == ID_function_application) {
        auto &app = to_function_application_expr(expr);

        //irep_idt s = ID_symbol;

        if (app.function().id() == ID_symbol) {
            // look up the symbol
            auto func = to_symbol_expr(app.function());
            auto f_it = problem.defined_functions.find(func);

            if (f_it != problem.defined_functions.end()) {
                // Does it have a definition? It's otherwise uninterpreted.
                if (!f_it->second.is_nil()) {
                    exprt body = f_it->second;

                    if (body.id() == ID_lambda) {
                        body = to_lambda_expr(body).application(app.arguments());
                    }
                    expand_function_applications(body, problem); // rec. call
                    expr = body;
                }
            }
        }
    }
}

smt_problemt expand_let_expressions(smt_problemt& problem) {
    for (auto &x: problem.assertions) {
        expand_let_expressions(x);
    }
    return problem;
}


smt_problemt generate_valid_problem(smt_problemt problem, namespacet& ns, messaget& message) {

    decision_proceduret::resultt res = solve_problem(problem, ns, message);

    smt_problemt new_valid_problem;
    if (res == decision_proceduret::resultt::D_SATISFIABLE) {

        // check for validity
        smt_problemt neg_problem = negate_problem(problem);
        decision_proceduret::resultt new_res = solve_problem(neg_problem, ns, message);

        if (new_res == decision_proceduret::resultt::D_UNSATISFIABLE) { // original problem is valid
            new_valid_problem = problem;
        } else if (new_res == decision_proceduret::resultt::D_ERROR) {
            throw solver_timeout("SMT solver error.");
        } else {
            // replace the free variables in the assertions with the values from the model
            new_valid_problem = substitute_model_into_problem(problem);
            // print the new smt_problem
            message.debug() << "Problem is satisfiable, working with substitution" << messaget::eom;
        }
    } else if (res == decision_proceduret::resultt::D_UNSATISFIABLE) {
        message.debug() << "Problem is UNSAT, working with negation." << messaget::eom;
        new_valid_problem = negate_problem(problem);
    } else if (res == decision_proceduret::resultt::D_ERROR) {
        throw solver_timeout("SMT solver error.");
    }

    return new_valid_problem;

}

// send the problem to an SMT solver and get the result
decision_proceduret::resultt solve_problem(smt_problemt &problem, namespacet &ns, messaget &message) {
    message.debug() << "Solving problem with smt2 solver" << messaget::eom;
    smt2_dect solver(
            ns, "fastsynth", "generated by fastsynth",
            problem.logic, smt2_dect::solvert::Z3, message.get_message_handler(), {"-T:"+std::to_string(SMT_SOLVER_TIMEOUT)});

    for (const auto &a: problem.assertions) {
        exprt copy_of_a = a;
        expand_function_applications(copy_of_a, problem);
        solver.set_to_true(copy_of_a);
    }

    decision_proceduret::resultt result = solver();

    switch (result) {
        case decision_proceduret::resultt::D_SATISFIABLE: {
            message.debug() << "RESULT is sat" << messaget::eom;
            for (const auto &e: problem.free_var) {
                exprt val = solver.get(e.first);
                problem.free_var[e.first] = val;
            }
            return result;
        }
        case decision_proceduret::resultt::D_UNSATISFIABLE: {
            message.debug() << "RESULT is unsat" << messaget::eom;
            return result;
        }

        case decision_proceduret::resultt::D_ERROR:
        default: {
            message.debug() << "ERROR from smt solver " << messaget::eom;
            return result;
        }
    }
    // UNREACHABLE
    return decision_proceduret::resultt::D_ERROR;
}

smt_problemt substitute_model_into_problem(const smt_problemt &problem) {
    smt_problemt new_problem;
    for (const auto &a: problem.assertions) {
        replace_symbolt replace_symbol;
        for (const auto &var: problem.free_var) {
            if (var.second.is_not_nil()) {
                replace_symbol.insert(var.first, var.second);
            }
        }

        exprt new_assertion = a;
        replace_symbol(new_assertion);
        new_problem.assertions.push_back(new_assertion);
    }

    sort(new_problem.assertions.begin(), new_problem.assertions.end());
    new_problem.assertions.erase(std::unique(new_problem.assertions.begin(), new_problem.assertions.end()),
                                 new_problem.assertions.end());

    new_problem.logic = problem.logic;
    new_problem.comments = problem.comments;
    new_problem.filename = problem.filename;

    for (const auto &f: problem.defined_functions) {
        new_problem.defined_functions[f.first] = f.second;
    }

    return new_problem;
}

void test_cvc5(message_handlert &mh)
{
    std::cout<<"making cvc5 problem"<<std::endl;
    messaget message(mh);
    // construct problem
    sygus_problemt problem;
    problem.logic="LIA";
    problem.free_var.push_back(symbol_exprt("x", integer_typet()));
    problem.free_var.push_back(symbol_exprt("y", integer_typet()));
    // construct synth fun command
    synth_fun_commandt synth_fun;
    synth_fun.id = "f";
    std::vector<typet> domain;
    domain.push_back(integer_typet());
    domain.push_back(integer_typet());
    synth_fun.type=mathematical_function_typet(domain, integer_typet());
    std::vector<irep_idt> params;
    params.push_back("a");
    params.push_back("b");
    synth_fun.parameters=params;
    problem.synth_fun=synth_fun;
    // add constraints
    std::vector<exprt> inputs;
    inputs.push_back(symbol_exprt("x", integer_typet()));
    inputs.push_back(symbol_exprt("y", integer_typet()));
    problem.assertions.push_back(
        equal_exprt(create_func_app("f", inputs, integer_typet()), 
                    plus_exprt(inputs[0], inputs[1])));
    cvc5_syntht cvc5_synth(mh, 10000);
    
    if(cvc5_synth(problem)==decision_proceduret::resultt::D_SATISFIABLE)
    {
        std::cout << "Synthesis succeeded" << std::endl;
        for(const auto &f: cvc5_synth.get_solution())
        {
            std::cout<<format(f.first)<<" = "<<format(f.second)<<std::endl;
        }
    }
}


int smt2_frontend(const cmdlinet &cmdline) {

    console_message_handlert message_handler;
    messaget message(message_handler);

    // this is our default verbosity
    unsigned int v = messaget::M_STATISTICS;

    if (cmdline.isset("verbosity")) {
        v = std::stol(
                cmdline.get_value("verbosity"));;
        if (v > 10) {
            v = 10;
        }
    }

    if(cmdline.isset("test-cvc5"))
    {
        test_cvc5(message.get_message_handler());
        return 0;
    }

    // parse input file
    assert(cmdline.args.size() == 1);
    std::ifstream in(cmdline.args.front());

    if (!in) {
        std::cerr << "Failed to open input file" << std::endl;
        return 10;
    }



    

    symbol_tablet symbol_table;
    namespacet ns(symbol_table);

    message_handler.set_verbosity(v);
    parsert parser(in);
    // parse the problem
    try {
        parser.parse();
    } catch (const parsert::smt2_errort &e) {
        message.error() << e.get_line_no() << ": "
                        << e.what() << messaget::eom;
        return 20;
    }

    smt_problemt problem = build_problem(parser);

    print_problem(problem, message.status());
    message.status() << messaget::eom;
    test_sygus_printing(problem, message.status());
    message.status() << messaget::eom;
    // solve the problem with an smt solver

    decision_proceduret::resultt res = solve_problem(problem, ns, message);
    // print problem and model
    message.status() << "Solving with SMT solver:" << messaget::eom;

    if (res == decision_proceduret::resultt::D_SATISFIABLE) {
        // replace the free variables in the assertions with the values from the model
        message.status()<<"Problem is SATISFIABLE\n";
        print_model(problem, message.status());
        message.status() << messaget::eom;

    } else if (res == decision_proceduret::resultt::D_UNSATISFIABLE) {
        message.status()<<"Problem is UNSATISFIABLE\n";
        message.status() << messaget::eom;
    }
    else
    {
        message.status()<<"SMT result error/unknown/timeout\n";
        message.status() << messaget::eom;
    }

    message.status() << messaget::eom; // flush

    return 1;
}